#!/usr/bin/env python
import pygame
import random
import os.path

TILE = 16
COUNT_W = 16
COUNT_H = 16
SIZE_W = COUNT_W * TILE
SIZE_H = COUNT_H * TILE
SCALE = 3

DOWN = 0
LEFT = 1
RIGHT = 2
UP = 3


'''
class Dirt(pygame.sprite.Sprite):
	def __init__(self):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load('data/dirt.bmp')
		self.data = []
		for i in range(0, COUNT_W * COUNT_H):	
			self.data.append(5)
	
	def get_data(self, (x, y)):
		return self.data[y * COUNT_H + x]

	def set_data(self, (x, y), value):
		self.data[y * COUNT_H + x] = value
	
	def get_rect(self, (x, y), (w, h)):
		x = max(0, min(x, SIZE_W - w))
		y = max(0, min(y, SIZE_H - h))

		for i in range
	
	def update(self, screen, objects, timer):
		for i in range(0, len(self.data)):
			if self.data[i]:
				pos = (i % COUNT_W * TILE, i / COUNT_H * TILE)
				screen.blit(self.image, pos)
'''


class SoundManager():
	def __init__(self):
		self.death = pygame.mixer.Sound(os.path.join('data', 'sounds', 'death.wav'))
		self.dig = pygame.mixer.Sound(os.path.join('data', 'sounds', 'dig.wav'))
		self.dig.set_volume(0.6)
		self.escape = pygame.mixer.Sound(os.path.join('data', 'sounds', 'escape.wav'))
		self.mine = pygame.mixer.Sound(os.path.join('data', 'sounds', 'mine.wav'))
		self.spawn = pygame.mixer.Sound(os.path.join('data', 'sounds', 'spawn.wav'))
		self.spawn2 = pygame.mixer.Sound(os.path.join('data', 'sounds', 'spawn2.wav'))
	

class Dirt(pygame.sprite.Sprite):
	def __init__(self, (x, y)):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load(os.path.join('data', 'dirt.bmp'))
		self.image.set_colorkey((0, 128, 128))
		self.rect = pygame.Rect((x, y), (TILE, TILE))
		self.health = 4
	
	def update(self, screen, objects, timer):
		if self.health <= 0:
			self.kill()
		f = 4 - self.health
		rect = pygame.Rect((f * TILE, 0), (TILE, TILE))
		screen.blit(self.image, self.rect, rect)


class Rock(pygame.sprite.Sprite):
	def __init__(self, (x, y)):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load(os.path.join('data', 'rock.bmp'))
		self.image.set_colorkey((0, 128, 128))
		self.rect = pygame.Rect((x, y), self.image.get_size())
	
	def update(self, screen, objects, timer):
		screen.blit(self.image, self.rect)


class Exit(pygame.sprite.Sprite):
	def __init__(self, (x, y)):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load(os.path.join('data', 'exit.bmp'))
		self.image.set_colorkey((0, 128, 128))
		self.rect = pygame.Rect((x, y), self.image.get_size())
	
	def update(self, screen, objects, timer):
		screen.blit(self.image, self.rect)


class SnakeKing(pygame.sprite.Sprite):
	def __init__(self):
		pygame.sprite.Sprite.__init__(self)
		self.click = False
		self.ammo = 0
		self.dirt = 0
	
	def update(self, screen, objects, timer, sm):
		if self.dirt >= 20:
			self.ammo += 1
			self.dirt = 0

		clicked = pygame.mouse.get_pressed()

		if clicked[0]:
			x = pygame.mouse.get_pos()[0] / SCALE
			y = pygame.mouse.get_pos()[1] / SCALE
			rect = pygame.Rect((x - 4, y - 4), (TILE / 2, TILE / 2))

			for sprite in objects['dirt'].sprites():
				if sprite.rect.collidepoint((x, y)):
					if not (timer + 2) % 8:
						sm.dig.play()
					self.click = True
					sprite.health -= 1
					self.dirt += 1

			if not self.click and self.ammo:
				can_spawn = True
				for sprite in objects['dirt'].sprites():
					if sprite.rect.colliderect(rect):
						screen.fill((128, 0, 0), sprite.rect)
						can_spawn = False

				for sprite in objects['rock'].sprites():
					if sprite.rect.colliderect(rect):
						screen.fill((128, 0, 0), sprite.rect)
						can_spawn = False

				for sprite in objects['dirt'].sprites():
					if sprite.rect.colliderect(rect):
						screen.fill((128, 0, 0), sprite.rect)
						can_spawn = False

				mole = objects['mole']
				mole_rect = pygame.Rect((mole.rect.x - 4 - TILE, mole.rect.y - 8 - TILE), (TILE * 3, TILE * 3))
				if mole_rect.colliderect(rect):
					screen.fill((128, 0, 0), mole_rect)
					can_spawn = False

				if can_spawn:
					sm.spawn.play()
					objects['snake'].add(Snake((x - 4, y - 6)))
					self.ammo -= 1
				self.click = True
		else:
			self.click = False


class Snake(pygame.sprite.Sprite):
	def __init__(self, (x, y)):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load(os.path.join('data', 'snake.bmp'))
		self.image.set_colorkey((0, 128, 128))
		self.rect = pygame.Rect((x, y), (TILE / 2, TILE / 2))
		self.x_vel = 0
		self.y_vel = 0
		self.frame = 0
		self.direction = random.randint(0, 3)
	
	def update(self, screen, objects, timer, sm):
		speed = 2
		self.x_vel = self.y_vel = 0
		if self.direction == DOWN:
			self.y_vel = speed
		elif self.direction == LEFT:
			self.x_vel = -speed
		elif self.direction == RIGHT:
			self.x_vel = speed
		elif self.direction == UP:
			self.y_vel = -speed

		self.rect.move_ip(self.x_vel, self.y_vel)
		if pygame.sprite.spritecollide(self, objects['rock'], False) or pygame.sprite.spritecollide(self, objects['dirt'], False) or not screen.get_rect().contains(self.rect):
			self.rect.move_ip(-self.x_vel, -self.y_vel)
			self.direction = random.randint(0, 3)

		if pygame.sprite.spritecollide(self, objects['mine'], True):
			for i in range(4):
				pos = (self.rect.x + random.randint(-8, 8), self.rect.y + random.randint(-8, 8))
				objects['effect'].add(Explosion(pos))
			sm.mine.play()
			self.kill()

		if not timer % 3:
			self.frame += 1
			if self.frame > 3:
				self.frame = 0

		if not timer % 90:
			self.direction = random.randint(0, 3)

		pos = (self.rect.x - 4, self.rect.y - 8)
		rect = pygame.Rect((self.frame * TILE, self.direction * TILE), (TILE, TILE))
		screen.blit(self.image, pos, rect)


class Mole(pygame.sprite.Sprite):
	def __init__(self, (x, y)):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load(os.path.join('data', 'mole.bmp'))
		self.image.set_colorkey((0, 128, 128))
		self.rect = pygame.Rect((x, y), (TILE / 2, TILE / 2))
		self.x_vel = 0
		self.y_vel = 0
		self.frame = 0
		self.is_moving = False
		self.space_pressed = False
		self.direction = DOWN
		self.alive = True
		self.exit = False
		self.ammo = 0
		self.dirt = 0
	
	def handle_events(self, objects, sm):
		pressed = pygame.key.get_pressed()
		self.x_vel = 2 * (pressed[pygame.K_RIGHT] - pressed[pygame.K_LEFT])
		self.y_vel = 2 * (pressed[pygame.K_DOWN] - pressed[pygame.K_UP])

		self.is_moving = self.x_vel or self.y_vel
		if self.y_vel: self.x_vel = 0

		if self.y_vel > 0: self.direction = DOWN
		if self.x_vel < 0: self.direction = LEFT
		if self.x_vel > 0: self.direction = RIGHT
		if self.y_vel < 0: self.direction = UP

		if pressed[pygame.K_SPACE] and self.ammo:
			if not self.space_pressed:
				sm.spawn2.play()
				objects['mine'].add(Mine(self.rect.topleft))
				self.ammo -= 1
				self.space_pressed = True
		else:
			self.space_pressed = False
	
	def update(self, screen, objects, timer, sm):
		self.handle_events(objects, sm)

		# Handle ammo
		if self.dirt >= 20:
			self.ammo += 1
			self.dirt = 0

		# Moves and collisions
		collide = False
		self.rect.move_ip(self.x_vel, self.y_vel)
		for sprite in objects['dirt']:
			if sprite.rect.colliderect(self.rect):
				if not timer % 4:
					sm.dig.play()
				sprite.health -= 1
				self.dirt += 1
				collide = True

		if pygame.sprite.spritecollide(self, objects['rock'], False) or not screen.get_rect().contains(self.rect) or collide:
			self.rect.move_ip(-self.x_vel, -self.y_vel)

		# Objects interactions
		if pygame.sprite.spritecollide(self, objects['snake'], False) and not self.exit:
			sm.death.play()
			objects['effect'].add(Ghost((self.rect.x - 4, self.rect.y - 4)))
			self.alive = False

		if objects['exit'].rect.colliderect(self.rect):
			sm.escape.play()
			self.exit = True

		# Blit
		if self.is_moving:
			if not timer % 3:
				self.frame += 1
				if self.frame > 3:
					self.frame = 0
		else:
			self.frame = 0

		pos = (self.rect.x - 4, self.rect.y - 8)
		rect = pygame.Rect((self.frame * TILE, self.direction * TILE), (TILE, TILE))
		screen.blit(self.image, pos, rect)


class Mine(pygame.sprite.Sprite):
	def __init__(self, (x, y)):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load(os.path.join('data', 'mine.bmp'))
		self.image.set_colorkey((0, 128, 128))
		self.rect = pygame.Rect((x, y), self.image.get_size())
	
	def update(self, screen, objects, timer):
		screen.blit(self.image, self.rect)


class Explosion(pygame.sprite.Sprite):
	def __init__(self, (x, y)):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load(os.path.join('data', 'explosion.bmp'))
		self.image.set_colorkey((0, 128, 128))
		self.rect = pygame.Rect((x, y), self.image.get_size())
		self.frame = 0
	
	def update(self, screen, objects, timer):
		self.frame += 1
		if self.frame > 3:
			self.kill()

		rect = pygame.Rect((self.frame * TILE, 0), (TILE, TILE))
		screen.blit(self.image, self.rect, rect)


class Ghost(pygame.sprite.Sprite):
	def __init__(self, (x, y)):
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.image.load(os.path.join('data', 'mole_ghost.bmp'))
		self.image.set_colorkey((0, 128, 128))
		self.image.set_alpha(192)
		self.rect = pygame.Rect((x, y), self.image.get_size())
		self.frame = 0
	
	def update(self, screen, objects, timer):
		if not timer % 4:
			self.frame = not(self.frame)

		self.rect.y -= 1

		rect = pygame.Rect((self.frame * TILE, 0), (TILE, TILE))
		screen.blit(self.image, self.rect, rect)


def load_level(name, objects):
	objects['rock'].empty()
	objects['snake'].empty()
	objects['mine'].empty()
	objects['effect'].empty()
	level = pygame.image.load(name)
	for y in range(level.get_height()):
		for x in range(level.get_width()):
			c = level.get_at((x, y))
			if c == (0, 0, 0):
				objects['rock'].add(Rock((x * TILE, y * TILE)))
			elif c == (0, 0, 255):
				objects['mole'] = Mole((x * TILE, y * TILE))

			if c != (0, 0, 255):
				objects['dirt'].add(Dirt((x * TILE, y * TILE)))
	exit = False
	while not exit:
		x = random.randint(0, COUNT_W)
		y = random.randint(0, COUNT_H)
		vx = x * TILE + TILE / 2
		vy = y * TILE + TILE / 2
		collide = False

		if objects['mole'].rect.collidepoint((vx, vy)):
			collide = True

		for sprite in objects['rock']:
			if sprite.rect.collidepoint((vx, vy)):
				collide = True

		if not collide:
			objects['exit'] = Exit((x * TILE, y * TILE))
			exit = True
	
	objects['king'] = SnakeKing()
#	objects['dirt'] = Dirt()


def main():
	pygame.mixer.pre_init(44100, -16, 8, 1024)
	pygame.init()

	window = pygame.display.set_mode((SIZE_W * SCALE, SIZE_H * SCALE))
	screen = pygame.Surface((SIZE_W, SIZE_H))
	clock = pygame.time.Clock()
	timer = 0
	run = True
	game_over = False

	sm = SoundManager()

	ground_image = pygame.image.load(os.path.join('data', 'sand.bmp'))
	ground = pygame.Surface((SIZE_W, SIZE_H))
	for i in range(COUNT_W * COUNT_H):
		ground.blit(ground_image, (i % COUNT_W * TILE, i / COUNT_H * TILE))
	
	cursor = pygame.image.load(os.path.join('data', 'cursor.bmp'))
	cursor.set_colorkey((0, 128, 128))
	pygame.mouse.set_visible(False)

	mine_icon = pygame.image.load(os.path.join('data', 'mine_icon.bmp'))
	mine_icon.set_colorkey((0, 128, 128))

	snake_icon = pygame.image.load(os.path.join('data', 'snake_icon.bmp'))
	snake_icon.set_colorkey((0, 128, 128))

	point_icon = pygame.image.load(os.path.join('data', 'point_icon.bmp'))
	point_icon.set_colorkey((0, 128, 128))

	font_large = pygame.font.Font(None, 32)
	font_small = pygame.font.Font(None, 16)

	rock = pygame.sprite.Group()
	dirt_group = pygame.sprite.Group()
	snake_group = pygame.sprite.Group()
	mine_group = pygame.sprite.Group()
	effect_group = pygame.sprite.Group()
	objects = {'rock': rock, 'dirt': dirt_group, 'mine': mine_group, 'snake': snake_group, 'effect': effect_group}
	load_level('levels/level0.bmp', objects)
	level_id = 0
	score_mole = 0
	score_king = 0
	winner = ''

	while run:
		clock.tick(30)
		timer += 1

		caption = "The Curse of the Snake King [FPS:%.2f]" % clock.get_fps()
		pygame.display.set_caption(caption)

		if not game_over:
			if not objects['mole'].alive:
				score_king += 1
				winner = 'king'
				game_over = True
			elif objects['mole'].exit:
				score_mole += 1
				winner = 'mole'
				game_over = True

		for event in pygame.event.get():
			if event.type == pygame.QUIT or (event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE):
				run = False
			if game_over and event.type == pygame.KEYDOWN and event.key == pygame.K_RETURN:
				level_id += 1
				if score_mole >= 3 or score_king >= 3:
					score_mole = 0
					score_king = 0
					level_id = 0
				name = 'levels/level%i.bmp' % level_id
				load_level(name, objects)
				game_over = False

		screen.blit(ground, (0, 0))
		objects['exit'].update(screen, objects, timer)
		objects['rock'].update(screen, objects, timer)
		objects['dirt'].update(screen, objects, timer)
		if not game_over:
			objects['king'].update(screen, objects, timer, sm)
		objects['mine'].update(screen, objects, timer)
		objects['snake'].update(screen, objects, timer, sm)
		if not game_over:
			objects['mole'].update(screen, objects, timer, sm)
		objects['effect'].update(screen, objects, timer)

		for i in range(objects['mole'].ammo):
			pos = ((4, 4 + 12 * i))
			screen.blit(mine_icon, pos)

		for i in range(objects['king'].ammo):
			pos = ((SIZE_W - 12, 4 + 12 * i))
			screen.blit(snake_icon, pos)

		rect = ((SIZE_W / 2 - 2, 11), (4, 2))
		screen.fill((128, 48, 0), rect)

		for i in range(3):
			pos = ((SIZE_W / 2 - 16 - 10 * i, 8))
			off = 0
			if score_mole > i:
				off = 1
			rect = ((off * 8, 0), (8, 8))
			screen.blit(point_icon, pos, rect)

		for i in range(3):
			pos = ((SIZE_W / 2 + 8 + 10 * i, 8))
			off = 0
			if score_king > i:
				off = 1
			rect = ((off * 8, 0), (8, 8))
			screen.blit(point_icon, pos, rect)

		if game_over:
			end = False
			if winner is 'mole':
				if score_mole < 3:
					text = font_large.render("Mole scores!", False, (255, 255, 255))
				else:
					text = font_large.render("Mole wins!", False, (255, 255, 255))
					end = True
			else:
				if score_king < 3:
					text = font_large.render("Snake King scores!", False, (255, 255, 255))
				else:
					text = font_large.render("Snake King wins!", False, (255, 255, 255))
					end = True

			screen.blit(text, (SIZE_W / 2 - text.get_width() / 2, SIZE_H / 2 - text.get_height() / 2 - 4))

			if end:
				text = font_small.render("Press enter to start a new game", False, (192, 192, 192))
			else:
				text = font_small.render("Press enter to continue", False, (192, 192, 192))
			screen.blit(text, (SIZE_W / 2 - text.get_width() / 2, SIZE_H / 2 - text.get_height() / 2 + 16))

		screen.blit(cursor, (pygame.mouse.get_pos()[0] / SCALE, pygame.mouse.get_pos()[1] / SCALE))

		pygame.transform.scale(screen, (SIZE_W * SCALE, SIZE_H * SCALE), window)
		pygame.display.update()

	pygame.quit()


if __name__ == "__main__":
	main()
